/* $Id: aprspacket.h 446 2011-02-21 16:32:42Z oh2gve $
 *
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 *
 *
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

#ifndef APRSPACKET_H
#define APRSPACKET_H

// non-std stuff.
#include <fap.h>

// std stuff.
#include <string>
#include <vector>
#include <ctime>


/// Represents one APRS packet.
/**
 * All methods of class assume that libfap has been initialized beforehand.
 */
class AprsPacket
{
public:

	/// Own type for location, uses WGS84.
	struct Location {
		/// Longitude in degrees.
		double lon;
		/// Latitude in degrees.
		double lat;
	};

	// Static members.

	/// Constructs a packet from TNC-2-formatted string.
	/**
	 * The given packet is first fed to libfap with is_ax25 flag set. If
	 * parsing fails, it is tried again with is_ax25 cleared.
	 * \param packet Packet in TNC-2 format.
	 * \return A packet is always returned, use error() to check for success.
	 */
	static AprsPacket* fromTnc2Packet(std::string const& packet);

	/// Constructs a packet from AX.25 frame.
	/**
	 * The given packet is cut from first \r or \n and fed to libfap with
	 * is_ax25 flag set.
	 * \param frame Bytes of the frame as received from an AX.25 network
	 * interface.
	 * \return A packet is always returned, use error() to check for success.
	 */
	static AprsPacket* fromAx25Frame(std::string const& frame);

	/// Constructs a packet from KISS frame.
	/**
	 * Given frame is converted to TNC-2 using libfap's converter, cut from
	 * first \r or \n and then fed to The Parser with is_ax25 flag set.
	 * \param frame Bytes of the frame as received from KISS TNC.
	 * \return A packet is always returned, use error() to check for success.
	 */
	static AprsPacket* fromKissFrame(std::string const& frame);


	// Constructors and destructor.

	/// Constructs a packet from given details.
	/**
	 * Use error() to check for success.
	 */
	AprsPacket(std::string const& source, std::string const& target, std::vector<std::string> const& path, std::string const& body, bool const& is_ax25);

	/// Copy constructor.
	AprsPacket(AprsPacket const& orig);

	/// Destructor.
	~AprsPacket();


	// Getters

	/// Return the packet in TNC2-format.
	std::string getTnc2Packet() const;

	/// Return the packet as a KISS frame.
	/**
	 * If packet is invalid, empty string is returned.
	 */
	std::string getKissFrame() const;

	/// Return the packet in format ready to be used as a body of another packet.
	/**
	 * Returned string starts with correct APRS data type indicator '}'. If
	 * packet is invalid, an empty string is returned.
	 * \param owncall Callsign of the gating station (the one running this code).
	 * \param path Middle element of the packet's path, for example TCPIP* with APRS-IS.
	 */
	std::string getAsThirdParty(std::string const& owncall, std::string const& path) const;

	/// Get loggable packet.
	/**
	 * Unprintable chars are marked as <xx> where xx is the char's ascii value
	 * in hex. For erroneous packet the error message is returned, with packet
	 * appended to the end in parentheses.
	 */
	std::string getLogLine() const;


	/// Return the packet being carried in the body of this packet.
	/**
	 * \return NULL if no third party packet was being carried.
	 */
	AprsPacket* getCarriedThirdParty() const;

	/// Returns unixtime of the moment when this packet object was created.
	/**
	 * For copy constructed objects timestamp of the original object is returned.
	 */
	time_t getTimestamp() const;

	/// Returns AX.25-level source callsign.
	std::string getSource() const;

	/// Returns name of the tactical APRS target in this packet.
	/**
	 * If packet contains object or item, name of it is returned. In other
	 * cases the AX.25 level source callsign is returned.
	 */
	std::string getLabel() const;

	/// Get location carried in the packet, if there is one.
	/**
	 *  \return True if the packet has a location, false if not
	 */
	bool getLocation(Location& loc) const;

	/// Get message and destination carried in packet, if those are available.
	bool getMessage(std::string& message, std::string& destination) const;
	
	/// Returns packet body.
	std::string getBody() const;


	// Operators

	/// Tell the status of a packet.
	/**
	 * \param message Error message.
	 * \return True if packet has an error, false if it is valid.
	 */
	bool error(std::string& message) const;

	/// Tells wether the packet can be gated from a network to another or not.
	/**
	 * Defaults to true. False is returned, if any of the following is true:
	 * 1) path of the packet contains NOGATE
	 * 2) packet contains a third-party packet
	 * 3) packet contains a query
	 */
	bool canBeGated() const;

	/// Tells wether the packet can be gated to APRS-IS or not.
	/**
	 * Defaults to true. False is returned, if any of the following is true:
	 * 1) path of the packet contains NOGATE, RFONLY, TCPIP or TCPXX
	 * 2) packet contains a third-party packet
	 * 3) packet contains a query
	 */
	bool canBeGatedToAprsIs() const;

	/// Tell if packet is same as another one.
	/**
	 * \return True if source, target and body of the packets are identical. False otherwise.
	 */
	bool operator==(AprsPacket const& theOther) const;
	
	
	// Setters
	
	void setTimestamp(time_t const& time);
	
	void addToPath(std::vector<std::string> const& path);

private:

	/// Constructor for the static from-methods.
	AprsPacket(fap_packet_t* data, bool const& is_ax25);

	/// Constructs an errorneous packet.
	AprsPacket(std::string const& origPacket, std::string const& errorMessage);

	/// Error message. If empty, we have no error.
	std::string errorMessage_;

	/// Bytes of original packet or frame given.
	std::string origPacket_;

	/// Packet data.
	fap_packet_t* data_;

	/// Tells wether the packet can be sent via AX.25 network or not.
	bool is_ax25_;

	/// Unixtime of object generation.
	time_t timestamp_;

};

#endif // APRSPACKET_H
