/* $Id: dupefilter.cpp 430 2011-02-13 18:46:34Z oh2gve $
 * 
 * Copyright 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Own stuff.
#include "dupefilter.h"
#include "aprspacket.h"

// std stuff.
#include <list>
using std::list;
#include <map>
using std::multimap;
using std::pair;
using std::make_pair;
#include <string>
using std::string;
#include <ctime>

// wx stuff.
#include <wx/thread.h>



DupeFilter::DupeFilter(time_t const& dupetime):
wxThread(wxTHREAD_JOINABLE), mutex_(), keepgoing_(true), storage_(), catalog_(), dupetime_(dupetime)
{
}



DupeFilter::~DupeFilter()
{
	while ( !storage_.empty() )
	{
		delete storage_.front();
		storage_.pop_front();
	}
}



wxThread::ExitCode DupeFilter::Entry()
{
	// Main loop
	while ( keepGoing() )
	{
		// Main loop is run once in second.
		Sleep(1000);
		
		// Check for too old packets in storage, remove if found.
		mutex_.Lock();
		while ( !storage_.empty() && time(NULL) - storage_.back()->getTimestamp() > dupetime_ )
		{
			// Found an old packet, now let's look for pointers to it in the catalogue.
			pair< multimap<string const, AprsPacket const*>::iterator,
			      multimap<string const, AprsPacket const*>::iterator > packetRange = catalog_.equal_range(storage_.back()->getLabel());
			if ( packetRange.second != packetRange.first )
			{
				// Check packets in the found range for exact match.
				for ( multimap<string const, AprsPacket const*>::iterator i = packetRange.first; i != packetRange.second; ++i )
				{
					if ( *i->second == *storage_.back() )
					{
						// We found a pointer, let's remove it.
						catalog_.erase(i);
						break;
					}
				}
			}
			
			// Remove the actual packet too.
			delete storage_.back();
			storage_.pop_back();
		}
		
		mutex_.Unlock();
	}
	
	return NULL;
}



void DupeFilter::terminate()
{
	mutex_.Lock();
	keepgoing_ = false;
	mutex_.Unlock();
}



bool DupeFilter::isDupe(AprsPacket const* packet)
{
	mutex_.Lock();
	
	// Check if the new packet is in catalog.
	pair< multimap<string const, AprsPacket const*>::iterator,
	      multimap<string const, AprsPacket const*>::iterator > packetRange = catalog_.equal_range(packet->getLabel());
	if ( packetRange.second != packetRange.first )
	{
		// Check packets in the found range for exact match.
		for ( multimap<string const, AprsPacket const*>::iterator i = packetRange.first; i != packetRange.second; ++i )
		{
			if ( *i->second == *packet )
			{
				// Found it, it's a dupe.
				mutex_.Unlock();
				return true;
			}
		}
	}
	
	// Apparently not a dupe, let's save it in storage. First we look for correct position.
	list<AprsPacket const*>::iterator pos;
	for ( pos = storage_.begin();
	      pos != storage_.end() && (*pos)->getTimestamp() > packet->getTimestamp();
	      pos++ )
	{
	}
	
	// Create a copy and save pointers to it.
	AprsPacket const* tmp = new AprsPacket(*packet);
	storage_.insert(pos, tmp);
	catalog_.insert( make_pair(packet->getLabel(), tmp) );
	
	mutex_.Unlock();
	return false;
}



bool DupeFilter::keepGoing()
{
	mutex_.Lock();
	bool retval = keepgoing_;
	mutex_.Unlock();
	return retval;
}
