/* $Id: ax25port.cpp 453 2011-02-27 10:06:27Z oh2gve $
 *
 * Copyright 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Compile contents only if AX.25 support is enabled.
#ifdef HAVE_LIBAX25



// own stuff.
#include "ax25port.h"
#include "../cfg.h"
using Cfg::ConnectionCfg;
#include "../packetlog.h"
#include "../dupefilter.h"
#include "../aprspacket.h"

// non-std stuff.
#include <netax25/ax25.h>
#include <netax25/axconfig.h>
#include <netax25/axlib.h>
#include <sys/socket.h>
#include <netinet/if_ether.h>
#include <sys/types.h>
#include <netinet/in.h>

// std stuff.
#include <cerrno>
#include <string>
using std::string;
#include <vector>
using std::vector;

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/string.h>




Ax25Iface::Ax25Iface(ConnectionCfg const& cfg, char const& conId, wxString const& owncall, Packetlog* packetlog, DupeFilter* dupefilter):
RFCon(cfg, conId, static_cast<char const*>(owncall.ToAscii()), packetlog, dupefilter),
rxSocket_(-1), txSocket_(-1), initialized_(true), port_(cfg.device), ifaceName_(), owncall_(owncall)
{
	// First we initialize libax25.
	if ( ax25_config_load_ports() == 0 )
	{
		wxLogMessage(wxT(" ax25port(") + port_ + wxT("): no AX.25 ports available."));
		initialized_ = false;
	}
	
	// Get interface name.
	if ( initialized_ )
	{
		char* ifname = ax25_config_get_dev(port_.char_str());
		if ( ifname == NULL )
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to get name of network interface."));
			initialized_ = false;
		}
		else
		{
			ifaceName_ = wxString::FromAscii(ifname);
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): successfully detected iface name ") + ifaceName_ + wxT(" for the port."));
		}
	}
	
	// Get port's callsign.
	if ( initialized_ )
	{
		char* tmpcall = ax25_config_get_addr(port_.char_str());
		if ( tmpcall )
		{
			portCall_ = wxString::FromAscii(tmpcall);
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): successfully detected callsign of the port. It's ") + portCall_ + wxT("."));
		}
		else
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to get callsign of the port."));
			initialized_ = false;
		}
	}
	
	// Open rx socket.
	if ( initialized_ && cfg.state != Cfg::WRITEONLY )
	{
		// How does one use AX.25 UI sockets correctly? This might not be it, but it seems to work.
		// ETH_P_AX25 is(?) promiscuous mode, we need to check port from each received packet.
		if ( ( rxSocket_ = socket(PF_PACKET, SOCK_PACKET, htons(ETH_P_AX25)) ) == -1 )
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to open rx socket (") + wxString::FromAscii(strerror(errno)) + wxT(")."));
			initialized_ = false;
		}
		
		// We want non-blocking io.
		if ( fcntl(rxSocket_, F_SETFL, O_NONBLOCK) < 0 )
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to setup rx socket (") + wxString::FromAscii(strerror(errno)) + wxT(")."));
			initialized_ = false;
		}
	}

	// Open tx socket.
	if ( initialized_  && cfg.state != Cfg::READONLY )
	{
		// Create the socket.
		if ( (txSocket_ = socket(PF_AX25, SOCK_DGRAM, 0) ) < 0 )
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to create tx socket."));
			initialized_ = false;
		}
	
		// Prepare to bind the socket to iface.
		struct full_sockaddr_ax25 txIfaceAddr;
		int txIfaceAddrLen = 0;
		if ( (txIfaceAddrLen = ax25_aton(portCall_.ToAscii(), &txIfaceAddr)) < 0 )
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to setup tx socket (") + wxString::FromAscii(strerror(errno)) + wxT(")."));
			initialized_ = false;
		}
		
		// Bind to iface.
		if ( bind(txSocket_, (struct sockaddr *)&txIfaceAddr, txIfaceAddrLen) < 0 )
		{
			wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to bind tx socket to the port (") + wxString::FromAscii(strerror(errno)) + wxT(")."));
			initialized_ = false;
		}
	}
}


Ax25Iface::~Ax25Iface()
{
	close(rxSocket_);
	close(txSocket_);
}


void* Ax25Iface::Entry()
{
	// Tell main that we're alive.
	threadStarted();
		
	// Check how we did at constructor.
	if ( initialized_ )
	{
		wxLogMessage(wxT(" ax25port(") + port_ + wxT("): connection opened."));
	}
	else
	{
		weHasError();
	}
	startupDone();
	
	
	// Main loop of this thread.
	while ( keepGoing() )
	{
		// Check if we can do input.
		if ( rxSocket_ != -1 )
		{
			// Prepare to get data from socket.
			char buffer[8192];
			struct sockaddr sa;
			int asize = sizeof(sa);
			
			// Check for data.
			int bytesRead = recvfrom(rxSocket_, buffer, WXSIZEOF(buffer), 0, &sa, (socklen_t*)&asize);
			if ( bytesRead < 0 )
			{
				if ( errno != EAGAIN )
				{
					wxString msg = wxT(" ax25port(");
					msg += port_;
					msg += wxT("): failed to read from socket, error code ");
					msg << errno;
					msg += wxT(".");
					wxLogMessage(msg);
					weHasError();
					break;
				}
			}
			else if ( bytesRead > 0 )
			{
				// Check where the bytes are from, pick only our iface.
				if ( strcmp(sa.sa_data, ifaceName_.ToAscii()) == 0 )
				{
					// Try to parse the bytes into packet.
					AprsPacket* packet = AprsPacket::fromAx25Frame(string(buffer+1, bytesRead-1));
					// If successfull, send to main.
					if ( packet != NULL )
					{
						input(packet);
					}
				}
				else
				{
					// Not for us.
					wxLogMessage(wxT(" ax25port(") + port_ + wxT(") debug: packet from ") + wxString::FromAscii(sa.sa_data) + wxT(" discarded."));
				}
			}
		}
		
		// Then let's do tx if we can.
		if ( txSocket_ != -1 )
		{
			// Try to get data.
			AprsPacket const* packet = output();
			if ( packet != NULL )
			{
				// Get packet as string.
				string output = packet->getTnc2Packet();
				
				// Find destination and path and convert them into "TOCALL VIA DIGI DIGI" format.
				string path;
				string::size_type startpos = output.find('>');
				string::size_type stoppos = output.find(':');
				// Path must exists and have a positive length. Body must also have positive length.
				if ( startpos != string::npos && stoppos != string::npos && ++startpos < stoppos && stoppos < output.size()-1)
				{
					// Loop through and find the parts.
					string buf;
					bool tocallFound = false;
					bool pathBegun = false;
					for ( unsigned int i = startpos; i <= stoppos; ++i )
					{
						if ( output.at(i) == ',' || i == stoppos )
						{
							if ( !tocallFound )
							{
								path = buf;
								tocallFound = true;
							}
							else if ( !pathBegun )
							{
								path += " VIA " + buf;
								pathBegun = true;
							}
							else
							{
								path += " " + buf;
							}
							buf.clear();
						}
						else
						{
							buf += output.at(i);
						}
					}
				
					// Put findings in addr blob.
					full_sockaddr_ax25 txDstAddr;
					int txDstAddrLen = 0;
					
					if ( (txDstAddrLen = ax25_aton(path.c_str(), &txDstAddr)) < 0 )
					{
						wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed convert ") + wxT(OWN_TARGET) + wxT(" into network format."));
						weHasError();
					}

					// Get packet body.
					output = output.substr(stoppos+1);
					
					// Send it.
					if ( sendto(txSocket_, output.c_str(), output.length(), 0, (sockaddr*)&txDstAddr, txDstAddrLen) < 0 )
					{
						wxLogMessage(wxT(" ax25port(") + port_ + wxT("): failed to transmit (") + wxString::FromAscii(strerror(errno)) + wxT(")."));
						weHasError();
					}
				}
			}
			delete packet;
		}
		
		// Wait a little before next checks.
		Sleep(10);

	} // while (!TestDestroy())
	
	// Tell user that we're dying.
	wxLogMessage(wxT(" ax25port(") + port_ + wxT("): shutting down."));
	
	return NULL;
}

#endif // HAVE_LIBAX25
