/* $Id: serialport.h 454 2011-02-27 10:30:45Z oh2gve $
 * 
 * Copyright 2009, 2010, 2011 Tapio Aaltonen
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SERIALPORT_H
#define SERIALPORT_H

#include <string>
#ifdef CPPSERWIN
#	include <windows.h>
#endif


/// Represents a serial port device.
class Serialport
{
public:
	   
	/// Constructor.
	/**
	 * Supported timeout values: 0..1000, value is in ms.
	 * Lockfile name is usually /var/lock/LCK..devicename on POSIX. Not used on Windows.
	 * Supported baudrates: 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600
	 * Supported charsizes: 7, 8
	 * Supported parities: N, E, O
	 * Supported stopbit counts: 1, 2
	*/
#ifdef CPPSERWIN
	Serialport(std::string const& deviceFile, unsigned int const& timeout, unsigned int const& baudrate,
	           unsigned int const& charsize, char const& parity, unsigned int const& stopbits, bool const& rtscts);
#else
	Serialport(std::string const& deviceFile, std::string const& lockFile, unsigned int const& timeout, unsigned int const& baudrate,
	           unsigned int const& charsize, char const& parity, unsigned int const& stopbits, bool const& rtscts);
#endif
	
	/// Destructor.
	~Serialport();
	
	/// Writes given data to the port device.
	/**
	 * \return True if all data was written, false in other cases.
	 */
	bool transmit(std::string const& data);
	
	/// Reads bytes from the port.
	/**
	 * If there's data in port buffer or timeouts is 0, returns
	 * immediately. If timeout is > 0 and there's no data, return after
	 * some data has appeared in port buffer or when timeout occurs.
	 * \param data Your buffer for received data. This is always cleared before checking for data from the port.
	 * \param max Maximum number of bytes to read from the port. Defaults to 512 (the max value) if not given.
	 * \return True in case of data received or timeout, false if error happened.
	 */
	bool receive(std::string& data, unsigned int const max = 512);
	
	/// Check port status.
	bool operator!() const;
	
	/// Returns error message.
	std::string getErrorMessage() const;
	
private:
	
	/// Copy constructor is not to be used.
	Serialport(Serialport const&);

	/// Assignment operator is not to be used.
	Serialport& operator=(Serialport const&);
	
	
	/// Timeout value in ms.
	unsigned int timeout_;
	
	/// In case of error the message is stored here.
	std::string error_;
	
	/// File handle for the port.
#ifdef CPPSERWIN
	HANDLE portHandle_;
#else
	int portHandle_;
	
	/// Lock file name on *NIX.
	std::string lockFileName_;
#endif
};

#endif // SERIALPORT_H
